# Copyright 2020 toyou Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
acs5000 san for common driver
It will be called by iSCSI driver
"""

import json
import random

from eventlet import greenthread
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import units
import paramiko

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder import utils as cinder_utils
from cinder.volume import driver
from cinder.volume.drivers.san import san
from cinder.volume import volume_utils

VOLUME_PREFIX = 'cinder-'

LOG = logging.getLogger(__name__)
acs5000c_opts = [
    cfg.ListOpt(
        'acs5000_volpool_name',
        default=['pool01'],
        help='Comma separated list of storage system storage '
             'pools for volumes.'),
    cfg.IntOpt(
        'acs5000_copy_interval',
        default=5,
        min=3,
        max=100,
        help='When volume copy task is going on,refresh volume '
             'status interval')
]
CONF = cfg.CONF
CONF.register_opts(acs5000c_opts)


class Command(object):

    def __init__(self, run_ssh):
        self._ssh = run_ssh

    def _run_ssh(self, ssh_cmd):
        try:
            return self._ssh(ssh_cmd)
        except processutils.ProcessExecutionError as e:
            msg = (_('CLI Exception output:\n command: %(cmd)s\n '
                     'stdout: %(out)s\n stderr: %(err)s.') %
                   {'cmd': ssh_cmd,
                    'out': e.stdout,
                    'err': e.stderr})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def run_ssh_info(self, ssh_cmd, key=False):
        """Run an SSH command and return parsed output."""
        out, err = self._run_ssh(ssh_cmd)
        if len(err):
            msg = (_('Execute command %(cmd)s failed, '
                     'out: %(out)s, err: %(err)s.') %
                   {'cmd': ' '.join(ssh_cmd),
                    'out': str(out),
                    'err': str(err)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        try:
            info = json.loads(out)
        except json.JSONDecodeError as e:
            msg = (_('Parse response error from CLI command %(cmd)s, '
                     'out: %(out)s, err: %(err)s') %
                   {'cmd': ' '.join(ssh_cmd),
                    'out': str(out),
                    'err': e})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if not isinstance(info, dict):
            msg = (_('Unexpected format from CLI command %(cmd)s, '
                     'result: %(info)s.') %
                   {'cmd': ' '.join(ssh_cmd),
                    'info': str(info)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        require = ('key', 'msg', 'arr')
        require_valid = True
        for r in require:
            if r not in info.keys():
                require_valid = False
                break
        if not require_valid:
            msg = (_('Unexpected response from CLI command %(cmd)s, '
                     'require \'key\' \'msg\' \'arr\'. out: %(info)s.') %
                   {'cmd': ' '.join(ssh_cmd),
                    'info': str(info)})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        elif int(info['key']) != 0:
            msg = (_('Unexpected error output from CLI command %(cmd)s, '
                     'key: %(key)s, msg: %(msg)s.') %
                   {'cmd': ' '.join(ssh_cmd),
                    'msg': info['msg'],
                    'key': info['key']})
            LOG.error(msg)
            if not key:
                raise exception.VolumeBackendAPIException(data=msg)
        if key:
            info['key'] = int(info['key'])
            return info
        else:
            return info['arr']

    def get_system(self):
        ssh_cmd = ['cinder', 'Storage', 'sshGetSystem']
        return self.run_ssh_info(ssh_cmd)

    def get_ip_connect(self):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshGetIpConnect']
        return self.run_ssh_info(ssh_cmd)

    def get_pool_info(self, pool):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshGetPoolInfo',
                   '--poolName',
                   pool]
        return self.run_ssh_info(ssh_cmd)

    def get_volume(self, volume):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshGetVolume']
        if not volume:
            return []
        elif isinstance(volume, str):
            ssh_cmd.append('--name')
            ssh_cmd.append(volume)
        elif isinstance(volume, list):
            for vol in volume:
                ssh_cmd.append('--name')
                ssh_cmd.append(vol)
        result = self.run_ssh_info(ssh_cmd)
        if not result:
            return []
        else:
            return result

    def ls_ctr_info(self):
        ssh_cmd = ['cinder', 'Storage', 'sshGetCtrInfo']
        ctrs = self.run_ssh_info(ssh_cmd)
        nodes = {}
        for node_data in ctrs:
            nodes[node_data['id']] = {
                'id': node_data['id'],
                'name': node_data['name'],
                'iscsi_name': node_data['iscsi_name'],
                'WWNN': node_data['WWNN'],
                'WWPN': [],
                'status': node_data['status'],
                'ipv4': [],
                'ipv6': [],
                'enabled_protocols': []
            }
        return nodes

    def create_volume(self, name, size, pool_name, type='0'):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshCreateVolume',
                   '--volumesize',
                   size,
                   '--volumename',
                   name,
                   '--cinderPool',
                   pool_name,
                   '--type',
                   type]
        return self.run_ssh_info(ssh_cmd, key=True)

    def delete_volume(self, volume):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshDeleteVolume',
                   '--cinderVolume',
                   volume]
        return self.run_ssh_info(ssh_cmd)

    def extend_volume(self, volume, size):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshCinderExtendVolume',
                   '--cinderVolume',
                   volume,
                   '--extendunit',
                   'gb',
                   '--extendsize',
                   str(size)]
        return self.run_ssh_info(ssh_cmd, key=True)

    def create_clone(self, volume_name, clone_name):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshMkLocalClone',
                   '--cinderVolume',
                   volume_name,
                   '--cloneVolume',
                   clone_name]
        return self.run_ssh_info(ssh_cmd, key=True)

    def start_clone(self, volume_name, snapshot=''):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshMkStartLocalClone',
                   '--cinderVolume',
                   volume_name,
                   '--snapshot',
                   snapshot]
        return self.run_ssh_info(ssh_cmd, key=True)

    def delete_clone(self, volume_name, snapshot=''):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshRemoveLocalClone',
                   '--name',
                   volume_name,
                   '--snapshot',
                   snapshot]
        return self.run_ssh_info(ssh_cmd, key=True)

    def create_lun_map(self, volume_name, protocol, host):
        """Map volume to host."""
        LOG.debug('enter: create_lun_map volume %s.', volume_name)
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshMapVoltoHost',
                   '--cinderVolume',
                   volume_name,
                   '--protocol',
                   protocol]
        if isinstance(host, list):
            for ht in host:
                ssh_cmd.append('--host')
                ssh_cmd.append(ht)
        else:
            ssh_cmd.append('--host')
            ssh_cmd.append(str(host))
        return self.run_ssh_info(ssh_cmd, key=True)

    def delete_lun_map(self, volume_name, protocol, host):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshDeleteLunMap',
                   '--cinderVolume',
                   volume_name,
                   '--protocol',
                   protocol]
        if isinstance(host, list):
            for ht in host:
                ssh_cmd.append('--cinderHost')
                ssh_cmd.append(ht)
        else:
            ssh_cmd.append('--cinderHost')
            ssh_cmd.append(str(host))
        return self.run_ssh_info(ssh_cmd, key=True)

    def create_snapshot(self, volume_name, snapshot_name):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshCreateSnapshot',
                   '--volume',
                   volume_name,
                   '--snapshot',
                   snapshot_name]
        return self.run_ssh_info(ssh_cmd, key=True)

    def delete_snapshot(self, volume_name, snapshot_name):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshDeleteSnapshot',
                   '--volume',
                   volume_name,
                   '--snapshot',
                   snapshot_name]
        return self.run_ssh_info(ssh_cmd, key=True)

    def set_volume_property(self, name, setting):
        ssh_cmd = ['cinder',
                   'Storage',
                   'sshSetVolumeProperty',
                   '--volume',
                   name]
        for key, value in setting.items():
            ssh_cmd.extend(['--' + key, value])
        return self.run_ssh_info(ssh_cmd, key=True)


class Acs5000CommonDriver(san.SanDriver,
                          driver.MigrateVD,
                          driver.CloneableImageVD):
    """TOYOU ACS5000 storage abstract common class.

    .. code-block:: none

      Version history:
          1.0.0 - Initial driver

    """
    VENDOR = 'TOYOU'
    VERSION = '1.0.0'

    def __init__(self, *args, **kwargs):
        super(Acs5000CommonDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(acs5000c_opts)
        self._backend_name = self.configuration.safe_get('volume_backend_name')
        self.pools = self.configuration.acs5000_volpool_name
        self._cmd = Command(self._run_ssh)
        self.protocol = None
        self._state = {'storage_nodes': {},
                       'enabled_protocols': set(),
                       'system_name': None,
                       'system_id': None,
                       'code_level': None,
                       'version': None}

    @staticmethod
    def get_driver_options():
        additional_opts = driver.BaseVD._get_oslo_driver_opts(
            'san_ip', 'san_ssh_port', 'san_login', 'san_password',
            'ssh_conn_timeout', 'ssh_min_pool_conn', 'ssh_max_pool_conn')
        return acs5000c_opts + additional_opts

    @volume_utils.trace_method
    def do_setup(self, ctxt):
        """Check that we have all configuration details from the storage."""
        self._validate_pools_exist()

        self._state.update(self._cmd.get_system())

        self._state['storage_nodes'] = self._cmd.ls_ctr_info()
        ports = self._cmd.get_ip_connect()
        if len(ports) > 0:
            self._state['enabled_protocols'].add('iSCSI')
            for node in self._state['storage_nodes'].values():
                if node['id'] in ports.keys():
                    node['enabled_protocols'].append('iSCSI')
                    for port in ports[node['id']]:
                        node['ipv4'].append(port['ip'])
        return

    def _validate_pools_exist(self):
        LOG.debug('_validate_pools_exist. '
                  'pools: %s', ' '.join(self.pools))
        for pool in self.pools:
            pool_data = self._cmd.get_pool_info(pool)
            if not pool_data:
                msg = _('Failed getting details for pool %s.') % pool
                raise exception.InvalidInput(reason=msg)
        return True

    @volume_utils.trace_method
    def check_for_setup_error(self):
        """Ensure that the params are set properly."""
        if self._state['system_name'] is None:
            exception_msg = _('Unable to determine system name.')
            raise exception.VolumeBackendAPIException(data=exception_msg)
        if self._state['system_id'] is None:
            exception_msg = _('Unable to determine system id.')
            raise exception.VolumeBackendAPIException(data=exception_msg)
        if len(self._state['storage_nodes']) != 2:
            msg = _('do_setup: No configured nodes.')
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)
        if self.protocol not in self._state['enabled_protocols']:
            raise exception.InvalidInput(
                reason=(_('The storage device does not support %(prot)s. '
                          'Please configure the device to support %(prot)s '
                          'or switch to a driver using a different '
                          'protocol.') % {'prot': self.protocol}))
        required = ['san_ip',
                    'san_ssh_port',
                    'san_login',
                    'acs5000_volpool_name']
        for param in required:
            if not self.configuration.safe_get(param):
                raise exception.InvalidInput(
                    reason=_('%s is not set.') % param)

        if not self.configuration.san_password:
            raise exception.InvalidInput(
                reason='Password is required for authentication')

        return

    def _run_ssh(self, cmd_list, check_exit_code=True):
        cinder_utils.check_ssh_injection(cmd_list)
        command = ' '.join(cmd_list)
        if not self.sshpool:
            try:
                self.sshpool = self._set_up_sshpool(self.configuration.san_ip)
            except paramiko.SSHException as e:
                raise exception.VolumeDriverException(message=e)
        ssh_execute = self._ssh_execute(
            self.sshpool, command, check_exit_code)
        return ssh_execute

    def _set_up_sshpool(self, ip):
        port = self.configuration.get('san_ssh_port', 22)
        login = self.configuration.get('san_login')
        password = self.configuration.get('san_password')
        timeout = self.configuration.get('ssh_conn_timeout', 30)
        min_size = self.configuration.get('ssh_min_pool_conn', 1)
        max_size = self.configuration.get('ssh_max_pool_conn', 5)
        sshpool = ssh_utils.SSHPool(ip,
                                    port,
                                    timeout,
                                    login,
                                    password=password,
                                    min_size=min_size,
                                    max_size=max_size)
        return sshpool

    def _ssh_execute(
            self,
            sshpool,
            command,
            check_exit_code=True):
        # noinspection PyBroadException
        try:
            with sshpool.item() as ssh:
                try:
                    return processutils.ssh_execute(
                        ssh, command, check_exit_code=check_exit_code)
                except Exception as e:
                    LOG.error('Error has occurred: %s', e)
                    raise processutils.ProcessExecutionError(
                        exit_code=e.exit_code,
                        stdout=e.stdout,
                        stderr=e.stderr,
                        cmd=e.cmd)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error('Error running SSH command: %s', command)

    def create_volume(self, volume):
        LOG.debug('create_volume, volume %s.', volume['id'])
        volume_name = VOLUME_PREFIX + volume['id'][-12:]
        pool_name = volume_utils.extract_host(volume['host'], 'pool')
        ret = self._cmd.create_volume(
            volume_name,
            str(volume['size']),
            pool_name)
        if ret['key'] == 310:
            msg = _('Volume: %s with same name '
                    'already exists on the system.') % volume_name
            raise exception.VolumeBackendAPIException(data=msg)
        elif ret['key'] == 102:
            allow_size = 0
            for p in self._stats['pools']:
                if p['pool_name'] == pool_name:
                    allow_size = p['free_capacity_gb']
                    break
            raise exception.VolumeSizeExceedsLimit(size=int(volume['size']),
                                                   limit=allow_size)
        elif ret['key'] == 307:
            raise exception.VolumeLimitExceeded(allowed=96,
                                                name=volume_name)
        elif ret['key'] == 308:
            raise exception.VolumeLimitExceeded(allowed=4096,
                                                name=volume_name)
        model_update = None
        return model_update

    def delete_volume(self, volume):
        volume_name = VOLUME_PREFIX + volume['id'][-12:]
        self._cmd.delete_volume(volume_name)

    def create_snapshot(self, snapshot):
        volume_name = VOLUME_PREFIX + snapshot['volume_name'][-12:]
        snapshot_name = VOLUME_PREFIX + snapshot['name'][-12:]
        ret = self._cmd.create_snapshot(volume_name, snapshot_name)
        if ret['key'] == 303:
            raise exception.VolumeNotFound(volume_id=volume_name)
        elif ret['key'] == 503:
            raise exception.SnapshotLimitExceeded(allowed=4096)
        elif ret['key'] == 504:
            raise exception.SnapshotLimitExceeded(allowed=64)

    def delete_snapshot(self, snapshot):
        volume_name = VOLUME_PREFIX + snapshot['volume_name'][-12:]
        snapshot_name = VOLUME_PREFIX + snapshot['name'][-12:]
        ret = self._cmd.delete_snapshot(volume_name, snapshot_name)
        if ret['key'] == 505:
            raise exception.SnapshotNotFound(snapshot_id=snapshot['id'])

    def create_volume_from_snapshot(self, volume, snapshot):
        snapshot_name = VOLUME_PREFIX + snapshot['name'][-12:]
        volume_name = VOLUME_PREFIX + volume['id'][-12:]
        source_volume = VOLUME_PREFIX + snapshot['volume_name'][-12:]
        pool = volume_utils.extract_host(volume['host'], 'pool')
        self._cmd.create_volume(volume_name,
                                str(volume['size']),
                                pool, '10')
        self._local_clone_copy(source_volume,
                               volume_name,
                               'create_volume_from_snapshot',
                               snapshot_name)

    def create_cloned_volume(self, tgt_volume, src_volume):
        clone_name = VOLUME_PREFIX + tgt_volume['id'][-12:]
        volume_name = VOLUME_PREFIX + src_volume['id'][-12:]
        tgt_pool = volume_utils.extract_host(tgt_volume['host'], 'pool')
        try:
            self._cmd.create_volume(clone_name, str(
                tgt_volume['size']), tgt_pool, '10')
            self._local_clone_copy(
                volume_name, clone_name, 'create_cloned_volume')
        except exception.VolumeBackendAPIException:
            self._cmd.delete_volume(clone_name)
            raise exception.VolumeBackendAPIException(
                data='create_cloned_volume failed.')

    def extend_volume(self, volume, new_size):
        volume_name = VOLUME_PREFIX + volume['id'][-12:]
        ret = self._cmd.extend_volume(volume_name, int(new_size))
        if ret['key'] == 303:
            raise exception.VolumeNotFound(volume_id=volume_name)
        elif ret['key'] == 321:
            msg = _('Volume capacity shall not be '
                    'less than the current size %sG.') % volume['size']
            raise exception.VolumeBackendAPIException(data=msg)
        elif ret['key'] == 102:
            pool_name = volume_utils.extract_host(volume['host'], 'pool')
            allow_size = 0
            for p in self._stats['pools']:
                if p['pool_name'] == pool_name:
                    allow_size = p['free_capacity_gb']
                    break
            raise exception.VolumeSizeExceedsLimit(size=int(new_size),
                                                   limit=allow_size)

    def migrate_volume(self, ctxt, volume, host):
        LOG.debug('enter: migrate_volume id %(id)s, host %(host)s',
                  {'id': volume['id'], 'host': host['host']})
        pool = volume_utils.extract_host(volume['host'], 'pool')
        if 'system_id' not in host['capabilities']:
            LOG.error('Target host has no system_id')
            return (False, None)
        if host['capabilities']['system_id'] != self._state['system_id']:
            LOG.info('The target host does not belong to the same '
                     'storage system as the current volume')
            return (False, None)
        if host['capabilities']['pool_name'] == pool:
            LOG.info('The target host belongs to the same storage system '
                     'and pool as the current volume.')
            return (True, None)
        LOG.info('The target host belongs to the same storage system '
                 'as the current but to a different pool. '
                 'The same storage system will clone volume into the new pool')
        volume_name = VOLUME_PREFIX + volume['id'][-12:]
        tmp_name = VOLUME_PREFIX + 'tmp'
        tmp_name += str(random.randint(0, 999999)).zfill(8)
        self._cmd.create_volume(tmp_name,
                                str(volume['size']),
                                host['capabilities']['pool_name'],
                                '10')
        self._local_clone_copy(
            volume_name, tmp_name, 'migrate_volume')
        self._cmd.delete_volume(volume_name)
        self._cmd.set_volume_property(tmp_name,
                                      {'type': '"RAID Volume"',
                                       'new_name': volume_name})
        return (True, None)

    def get_volume_stats(self, refresh=False):
        """Get volume stats.

        If we haven't gotten stats yet or 'refresh' is True,
        run update the stats first.
        """
        if not self._stats or refresh:
            self._update_volume_stats()
        return self._stats

    def _update_volume_stats(self):
        """Retrieve stats info from volume group."""
        LOG.debug('Updating volume stats, '
                  'pools: \'%(host)s#%(pool)s\'.',
                  {'host': self.host,
                   'pool': ','.join(self.pools)})
        data = {}
        data['vendor_name'] = self.VENDOR
        data['driver_version'] = self.VERSION
        data['storage_protocol'] = self.protocol
        backend_name = self.configuration.safe_get('volume_backend_name')
        data['volume_backend_name'] = (backend_name or
                                       self._state['system_name'])
        data['pools'] = [self._build_pool_stats(pool)
                         for pool in self.pools]

        self._stats = data

    def _build_pool_stats(self, pool):
        """Build pool status"""
        pool_stats = {}
        try:
            pool_data = self._cmd.get_pool_info(pool)
            if pool_data:
                total_capacity_gb = float(pool_data['capacity']) / units.Gi
                free_capacity_gb = float(pool_data['free_capacity']) / units.Gi
                allocated_capacity_gb = float(
                    pool_data['used_capacity']) / units.Gi
                total_volumes = None
                if 'total_volumes' in pool_data.keys():
                    total_volumes = int(pool_data['total_volumes'])
                pool_stats = {
                    'pool_name': pool_data['name'],
                    'total_capacity_gb': total_capacity_gb,
                    'free_capacity_gb': free_capacity_gb,
                    'allocated_capacity_gb': allocated_capacity_gb,
                    'compression_support': True,
                    'reserved_percentage':
                        self.configuration.reserved_percentage,
                    'QoS_support': False,
                    'consistencygroup_support': False,
                    'multiattach': False,
                    'easytier_support': False,
                    'total_volumes': total_volumes,
                    'system_id': self._state['system_id']}
            else:
                msg = _('Backend storage pool "%s" not found.') % pool
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
        except exception.VolumeBackendAPIException:
            msg = _('Failed getting details for pool %s.') % pool
            raise exception.VolumeBackendAPIException(data=msg)

        return pool_stats

    def _local_clone_copy(self, volume, clone, action=None, snapshot=''):
        LOG.debug('enter: copy volume %(vol)s to %(clone)s by %(action)s.',
                  {'vol': volume,
                   'clone': clone,
                   'action': action})
        if self._wait_volume_copy(volume, clone, action, 'wait'):
            LOG.info('start copy task.')
            ret = self._cmd.create_clone(volume, clone)
            if ret['key'] != 0:
                self._cmd.delete_volume(clone)
            if ret['key'] == 306:
                raise exception.VolumeBackendAPIException(
                    data='The source volume must not be larger '
                         'than the target volume in a clone relation. ')
            elif ret['key'] == 0:
                ret = self._cmd.start_clone(volume, snapshot)
                if ret['key'] == 505:
                    raise exception.SnapshotNotFound(snapshot_id=snapshot)
        else:
            LOG.error('%(action)s failed.', {'action': action})
            raise exception.VolumeBackendAPIException(data='clone failed!')

        if self._wait_volume_copy(volume, clone, action, 'copy'):
            self._cmd.delete_clone(volume, snapshot)
            LOG.info('%s successfully.', action)
        else:
            LOG.error('%(action)s failed.', {'action': action})
            raise exception.VolumeBackendAPIException(data='clone failed!')
        LOG.debug('leave: copy volume %(vol)s to %(clone)s by %(action)s. ',
                  {'vol': volume,
                   'clone': clone,
                   'action': action})

    @coordination.synchronized('acs5000-copy-{volume}-task')
    def _wait_volume_copy(self, volume, clone, function=None, action=None):
        LOG.debug('_wait_volume_copy, volume %s.', volume)
        if volume is None or clone is None:
            LOG.error('volume parameter error.')
            return False
        ret = False
        while_exit = False
        rescan = 0
        interval = self.configuration.acs5000_copy_interval
        wait_status = (
            'Initiating',
            'Rebuilding',
            'Erasing',
            'Delayed rebuilding')
        # All status
        # {"Offline", "Online", "Initiating",
        # ###"Rebuilding", "Migrating", "Parity chking",
        # ###"Cloning", "Rolling back", "Parity chking",
        # ###"Replicating", "Erasing", "Moving", "Replacing",
        # "Reclaiming", "Delayed rebuilding", "Relocation", "N/A"};
        # All health
        # {"Optimal", "Degraded", "Deleted", "Missing", "Failed",
        #  "Partially optimal", "N/A"}
        while True:
            rescan += 1
            volume_info = self._cmd.get_volume([volume, clone])
            if len(volume_info) == 2:
                for vol in volume_info:
                    if vol['type'] == 'BACKUP':
                        if vol['health'] == 'Optimal' and (
                                vol['status'] in wait_status):
                            LOG.info('%(function)s %(action)s task: '
                                     'rescan %(scan)s times, clone %(clone)s '
                                     'need wait,status is %(status)s, '
                                     'health is %(health)s, '
                                     'process is %(process)s%%. ',
                                     {'function': function,
                                      'action': action,
                                      'scan': rescan,
                                      'clone': vol['name'],
                                      'status': vol['status'],
                                      'health': vol['health'],
                                      'process': vol['r']})
                    elif vol['status'] == 'Cloning':
                        LOG.info('%(function)s %(action)s task: '
                                 'rescan %(scan)s times,volume %(volume)s '
                                 'copy process %(process)s%%. ',
                                 {'function': function,
                                  'action': action,
                                  'scan': rescan,
                                  'volume': vol['name'],
                                  'process': vol['r']})
                    elif vol['status'] == 'Queued':
                        LOG.info('%(function)s %(action)s task: '
                                 'rescan %(scan)s times, '
                                 'volume %(volume)s is in the queue. ',
                                 {'function': function,
                                  'action': action, 'scan': rescan,
                                  'volume': vol['name']})
                    elif (vol['type'] == 'RAID Volume'
                          and vol['status'] == 'Online'):
                        ret = True
                        while_exit = True
                        LOG.info('%(function)s %(action)s task: '
                                 'rescan %(scan)s times,volume %(volume)s '
                                 'copy task completed,status is Online. ',
                                 {'function': function,
                                  'action': action,
                                  'scan': rescan,
                                  'volume': vol['name']})
                    elif (vol['health'] == 'Optimal'
                          and (vol['status'] in wait_status)):
                        LOG.info('%(function)s %(action)s task: '
                                 'rescan %(scan)s times,volume %(volume)s '
                                 'need wait, '
                                 'status is %(status)s,health is %(health)s, '
                                 'process is %(process)s%%. ',
                                 {'function': function,
                                  'action': action,
                                  'scan': rescan,
                                  'volume': vol['name'],
                                  'status': vol['status'],
                                  'health': vol['health'],
                                  'process': vol['r']})
                    else:
                        LOG.info('%(function)s %(action)s task: '
                                 'rescan %(scan)s times,volume %(volume)s '
                                 'is not normal, '
                                 'status %(status)s,health is %(health)s. ',
                                 {'function': function,
                                  'action': action,
                                  'scan': rescan,
                                  'volume': vol['name'],
                                  'status': vol['status'],
                                  'health': vol['health']})
                        while_exit = True
                        break
            elif len(volume_info) == 1:
                while_exit = True
                if volume_info[0]['name'] == volume:
                    LOG.info('%(function)s %(action)s task: '
                             'rescan %(scan)s times,clone %(clone)s '
                             'does not exist! ',
                             {'function': function,
                              'action': action,
                              'scan': rescan,
                              'clone': clone})
                else:
                    LOG.info('%(function)s %(action)s task: '
                             'rescan %(scan)s times,volume %(volume)s '
                             'does not exist! ',
                             {'function': function,
                              'action': action,
                              'scan': rescan,
                              'volume': volume})
            else:
                while_exit = True
                LOG.info('%(function)s %(action)s task: '
                         'rescan %(scan)s times,volume %(volume)s '
                         'clone %(clone)s does not exist! ',
                         {'function': function,
                          'action': action,
                          'scan': rescan,
                          'volume': volume,
                          'clone': clone})

            if while_exit:
                break
            greenthread.sleep(interval)
        return ret



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

import ddt

from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.v3 import messages
from cinder import context
from cinder import exception
from cinder.message import api as message_api
from cinder.message import message_field
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v3 import fakes as v3_fakes
from cinder.tests.unit import test


NS = '{http://docs.openstack.org/api/openstack-block-storage/3.0/content}'


@ddt.ddt
class MessageApiTest(test.TestCase):
    def setUp(self):
        super(MessageApiTest, self).setUp()
        self.ext_mgr = extensions.ExtensionManager()
        self.ext_mgr.extensions = {}
        self.controller = messages.MessagesController(self.ext_mgr)

        self.maxDiff = None
        self.ctxt = context.RequestContext('admin', 'fakeproject', True)

    def _expected_message_from_controller(self, id):
        message = v3_fakes.fake_message(id)
        links = [
            {'href': 'http://localhost/v3/fakeproject/messages/%s' % id,
             'rel': 'self'},
            {'href': 'http://localhost/fakeproject/messages/%s' % id,
             'rel': 'bookmark'},
        ]
        return {
            'message': {
                'id': message.get('id'),
                'user_message': "%s:%s" % (
                    message_field.translate_action(message.get('action_id')),
                    message_field.translate_detail(message.get('detail_id'))),
                'request_id': message.get('request_id'),
                'event_id': message.get('event_id'),
                'created_at': message.get('created_at'),
                'message_level': message.get('message_level'),
                'guaranteed_until': message.get('expires_at'),
                'links': links,
            }
        }

    def test_show(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        res_dict = self.controller.show(req, fakes.FAKE_UUID)

        ex = self._expected_message_from_controller(fakes.FAKE_UUID)
        self.assertEqual(ex, res_dict)

    def test_show_not_found(self):
        self.mock_object(message_api.API, 'get',
                         side_effect=exception.MessageNotFound(
                             message_id=fakes.FAKE_UUID))

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        self.assertRaises(exception.MessageNotFound, self.controller.show,
                          req, fakes.FAKE_UUID)

    def test_show_pre_microversion(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)

        req = fakes.HTTPRequest.blank('/v3/messages/%s' % fakes.FAKE_UUID,
                                      version=mv.BASE_VERSION)
        req.environ['cinder.context'] = self.ctxt

        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller.show, req, fakes.FAKE_UUID)

    def test_delete(self):
        self.mock_object(message_api.API, 'get', v3_fakes.fake_message_get)
        self.mock_object(message_api.API, 'delete')

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        resp = self.controller.delete(req, fakes.FAKE_UUID)

        self.assertEqual(HTTPStatus.NO_CONTENT, resp.status_int)
        self.assertTrue(message_api.API.delete.called)

    def test_delete_not_found(self):
        self.mock_object(message_api.API, 'get',
                         side_effect=exception.MessageNotFound(
                             message_id=fakes.FAKE_UUID))

        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)

        self.assertRaises(exception.MessageNotFound, self.controller.delete,
                          req, fakes.FAKE_UUID)

    @ddt.data(mv.get_prior_version(mv.RESOURCE_FILTER),
              mv.RESOURCE_FILTER, mv.LIKE_FILTER)
    @mock.patch('cinder.api.common.reject_invalid_filters')
    def test_message_list_with_general_filter(self, version, mock_update):
        url = '/v3/%s/messages' % fakes.FAKE_UUID
        req = fakes.HTTPRequest.blank(url,
                                      version=version,
                                      use_admin_context=False)
        self.controller.index(req)

        if version != mv.get_prior_version(mv.RESOURCE_FILTER):
            support_like = True if version == mv.LIKE_FILTER else False
            mock_update.assert_called_once_with(req.environ['cinder.context'],
                                                mock.ANY, 'message',
                                                support_like)

    def test_index(self):
        self.mock_object(message_api.API, 'get_all',
                         return_value=[v3_fakes.fake_message(fakes.FAKE_UUID)])
        req = fakes.HTTPRequest.blank(
            '/v3/messages/%s' % fakes.FAKE_UUID, version=mv.MESSAGES)
        req.environ['cinder.context'] = self.ctxt

        res_dict = self.controller.index(req)

        ex = self._expected_message_from_controller(fakes.FAKE_UUID)
        expected = {
            'messages': [ex['message']]
        }
        self.assertDictEqual(expected, res_dict)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from cinder import exception
from cinder.image import accelerator
from cinder.tests.unit import test


class TestAccelerators(test.TestCase):
    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = True)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = True)
    # Compress test, QAT and GZIP available
    def test_compress_img_prefer_qat_when_available(self,
                                                    mock_gzip_exist,
                                                    mock_qat_exist,
                                                    mock_exec):

        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.compress_img(run_as_root=True)

        expected = [
            mock.call('qzip', '-k', dest, '-o', dest,
                      run_as_root=True),
            mock.call('mv', dest + '.gz', dest,
                      run_as_root=True)
        ]

        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = False)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = True)
    # Compress test, QAT not available but GZIP available
    def test_compress_img_qat_accel_not_exist_gzip_exist(self,
                                                         mock_gzip_exist,
                                                         mock_qat_exist,
                                                         mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.compress_img(run_as_root=True)

        not_called = mock.call('qzip', '-k', dest, '-o', dest,
                               run_as_root=True)

        self.assertNotIn(not_called, mock_exec.call_args_list)

        expected = [
            mock.call('gzip', '-k', dest,
                      run_as_root=True),
            mock.call('mv', dest + '.gz', dest,
                      run_as_root=True)
        ]
        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = True)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = False)
    # Compress test, QAT available but GZIP not available
    def test_compress_img_prefer_qat_without_gzip(self,
                                                  mock_gzip_exist,
                                                  mock_qat_exist,
                                                  mock_exec):

        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.compress_img(run_as_root=True)

        expected = [
            mock.call('qzip', '-k', dest, '-o', dest,
                      run_as_root=True),
            mock.call('mv', dest + '.gz', dest,
                      run_as_root=True)
        ]

        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = False)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = False)
    # Compress test, no accelerator available
    def test_compress_img_no_accel_exist(self,
                                         mock_gzip_exist,
                                         mock_qat_exist,
                                         mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        self.assertRaises(exception.CinderException,
                          accelerator.ImageAccel,
                          source,
                          dest)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = True)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = True)
    # Decompress test, QAT and GZIP available
    def test_decompress_img_prefer_qat_when_available(self,
                                                      mock_gzip_exist,
                                                      mock_qat_exist,
                                                      mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.decompress_img(run_as_root=True)

        expected = [
            mock.call('mv', source, source + '.gz',
                      run_as_root=True),
            mock.call('qzip', '-d', source + '.gz',
                      run_as_root=True)
        ]

        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = False)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = True)
    # Decompress test, QAT not available but GZIP available
    def test_decompress_img_qat_accel_not_exist_gzip_exist(self,
                                                           mock_gzip_exist,
                                                           mock_qat_exist,
                                                           mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.decompress_img(run_as_root=True)

        not_called = mock.call('qzip', '-d', source + '.gz',
                               run_as_root=True)

        self.assertNotIn(not_called, mock_exec.call_args_list)

        expected = [
            mock.call('mv', source, source + '.gz',
                      run_as_root=True),
            mock.call('gzip', '-d', source + '.gz',
                      run_as_root=True)
        ]

        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = True)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = False)
    # Decompress test, QAT available but GZIP not available
    def test_decompress_img_prefer_qat_without_gzip(self,
                                                    mock_gzip_exist,
                                                    mock_qat_exist,
                                                    mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        accel = accelerator.ImageAccel(source, dest)
        accel.decompress_img(run_as_root=True)

        expected = [
            mock.call('mv', source, source + '.gz',
                      run_as_root=True),
            mock.call('qzip', '-d', source + '.gz',
                      run_as_root=True)
        ]

        mock_exec.assert_has_calls(expected)

    @mock.patch('cinder.utils.execute')
    @mock.patch('cinder.image.accelerators.qat.AccelQAT.is_accel_exist',
                return_value = False)
    @mock.patch('cinder.image.accelerators.gzip.AccelGZIP.is_accel_exist',
                return_value = False)
    # Decompress test, no accelerator available
    def test_decompress_img_no_accel_exist(self,
                                           mock_gzip_exist,
                                           mock_qat_exist,
                                           mock_exec):
        source = 'fake_path'
        dest = 'fake_path'

        self.assertRaises(exception.CinderException,
                          accelerator.ImageAccel,
                          source,
                          dest)



# Copyright 2016 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# product id is 2145 for SVC 6.1.0+. no product id for older version.
DEV_MODEL_SVC = '2145'
DEV_MODEL_STORWIZE = '2076'
DEV_MODEL_STORWIZE_V3500 = '2071'
DEV_MODEL_STORWIZE_V5000E = '2072'
DEV_MODEL_STORWIZE_V7000 = '2076'
DEV_MODEL_STORWIZE_V5000 = '2078'
DEV_MODEL_STORWIZE_V5000_1YR = '2077'
DEV_MODEL_FLASH_V9000 = '9846'
DEV_MODEL_FLEX = '4939'

REP_CAP_DEVS = (DEV_MODEL_SVC, DEV_MODEL_STORWIZE, DEV_MODEL_STORWIZE_V5000,
                DEV_MODEL_STORWIZE_V5000_1YR, DEV_MODEL_FLASH_V9000,
                DEV_MODEL_FLEX, DEV_MODEL_STORWIZE_V5000E)

# constants used for replication
GLOBAL = 'global'
METRO = 'metro'
GMCV = 'gmcv'
GMCV_MULTI = 'multi'
VALID_REP_TYPES = (GLOBAL, METRO, GMCV)
FAILBACK_VALUE = 'default'

DEFAULT_RC_TIMEOUT = 3600 * 24 * 7
DEFAULT_RC_INTERVAL = 5

DEFAULT_RCCG_TIMEOUT = 60 * 30
DEFAULT_RCCG_INTERVAL = 2

REPLICA_AUX_VOL_PREFIX = 'aux_'
REPLICA_CHG_VOL_PREFIX = 'chg_'

RCCG_PREFIX = 'rccg-'
HYPERCG_PREFIX = 'hycg-'

# remote mirror copy status
REP_CONSIS_SYNC = 'consistent_synchronized'
REP_CONSIS_COPYING = 'consistent_copying'
REP_CONSIS_STOP = 'consistent_stopped'
REP_SYNC = 'synchronized'
REP_IDL = 'idling'
REP_IDL_DISC = 'idling_disconnected'
REP_STATUS_ON_LINE = 'online'

# IOThrottling types
MBPS = 'mbps'
IOPS = 'iops'
IOPS_PER_GB = 'iops_per_gb'



#    Copyright 2015 Intel Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

import ddt
from oslo_utils import timeutils
import pytz

from cinder import exception
from cinder import objects
from cinder.tests.unit import fake_cluster
from cinder.tests.unit import fake_service
from cinder.tests.unit import objects as test_objects


@ddt.ddt
class TestService(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.sqlalchemy.api.service_get')
    def test_get_by_id(self, service_get):
        db_service = fake_service.fake_db_service()
        service_get.return_value = db_service
        service = objects.Service.get_by_id(self.context, 1)
        self._compare(self, db_service, service)
        service_get.assert_called_once_with(self.context, 1)

    @ddt.data(True, False)
    @mock.patch('cinder.db.service_get')
    def test_get_by_host_and_topic(self, show_disabled, service_get):
        db_service = fake_service.fake_db_service()
        service_get.return_value = db_service
        service = objects.Service.get_by_host_and_topic(
            self.context, 'fake-host', 'fake-topic', disabled=show_disabled)
        self._compare(self, db_service, service)
        service_get.assert_called_once_with(
            self.context, disabled=show_disabled, host='fake-host',
            topic='fake-topic')

    @mock.patch('cinder.db.service_get')
    def test_get_by_args(self, service_get):
        db_service = fake_service.fake_db_service()
        service_get.return_value = db_service
        service = objects.Service.get_by_args(
            self.context, 'fake-host', 'fake-key')
        self._compare(self, db_service, service)
        service_get.assert_called_once_with(
            self.context, host='fake-host', binary='fake-key')

    @mock.patch('cinder.db.service_create')
    def test_create(self, service_create):
        db_service = fake_service.fake_db_service()
        service_create.return_value = db_service
        service = objects.Service(context=self.context)
        service.create()
        self.assertEqual(db_service['id'], service.id)
        service_create.assert_called_once_with(self.context,
                                               {'uuid': mock.ANY})

    @mock.patch('cinder.db.service_update')
    def test_save(self, service_update):
        db_service = fake_service.fake_db_service()
        service = objects.Service._from_db_object(
            self.context, objects.Service(), db_service)
        service.topic = 'foobar'
        service.save()
        service_update.assert_called_once_with(self.context, service.id,
                                               {'topic': 'foobar'})

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.service_destroy')
    def test_destroy(self, service_destroy, utcnow_mock):
        service_destroy.return_value = {
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        db_service = fake_service.fake_db_service()
        service = objects.Service._from_db_object(
            self.context, objects.Service(), db_service)
        with mock.patch.object(service._context, 'elevated') as elevated_ctx:
            service.destroy()
            service_destroy.assert_called_once_with(elevated_ctx(), 123)
        self.assertTrue(service.deleted)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         service.deleted_at)

    @mock.patch('cinder.db.sqlalchemy.api.service_get')
    def test_refresh(self, service_get):
        db_service1 = fake_service.fake_db_service()
        db_service2 = db_service1.copy()
        db_service2['availability_zone'] = 'foobar'

        # On the second service_get, return the service with an updated
        # availability_zone
        service_get.side_effect = [db_service1, db_service2]
        service = objects.Service.get_by_id(self.context, 123)
        self._compare(self, db_service1, service)

        # availability_zone was updated, so a service refresh should have a
        # new value for that field
        service.refresh()
        self._compare(self, db_service2, service)
        service_get.assert_has_calls([mock.call(self.context, 123),
                                      mock.call.__bool__(),
                                      mock.call(self.context, 123)])

    @mock.patch('cinder.db.service_get_all')
    def test_get_minimum_version(self, service_get_all):
        services_update = [
            {'rpc_current_version': '1.0', 'object_current_version': '1.3'},
            {'rpc_current_version': '1.1', 'object_current_version': '1.2'},
            {'rpc_current_version': '2.0', 'object_current_version': '2.5'},
        ]
        expected = ('1.0', '1.2')
        services = [fake_service.fake_db_service(**s) for s in services_update]
        service_get_all.return_value = services

        min_rpc = objects.Service.get_minimum_rpc_version(self.context, 'foo')
        self.assertEqual(expected[0], min_rpc)
        min_obj = objects.Service.get_minimum_obj_version(self.context, 'foo')
        self.assertEqual(expected[1], min_obj)
        service_get_all.assert_has_calls(
            [mock.call(self.context, binary='foo', disabled=None)] * 2)

    @mock.patch('cinder.db.service_get_all')
    def test_get_minimum_version_liberty(self, service_get_all):
        services_update = [
            {'rpc_current_version': '1.0', 'object_current_version': '1.3'},
            {'rpc_current_version': '1.1', 'object_current_version': None},
            {'rpc_current_version': None, 'object_current_version': '2.5'},
        ]
        services = [fake_service.fake_db_service(**s) for s in services_update]
        service_get_all.return_value = services

        self.assertRaises(exception.ServiceTooOld,
                          objects.Service.get_minimum_rpc_version,
                          self.context, 'foo')
        self.assertRaises(exception.ServiceTooOld,
                          objects.Service.get_minimum_obj_version,
                          self.context, 'foo')

    @mock.patch('cinder.db.service_get_all')
    def test_get_minimum_version_no_binary(self, service_get_all):
        services_update = [
            {'rpc_current_version': '1.0', 'object_current_version': '1.3'},
            {'rpc_current_version': '1.1', 'object_current_version': '1.2'},
            {'rpc_current_version': '2.0', 'object_current_version': '2.5'},
        ]
        services = [fake_service.fake_db_service(**s) for s in services_update]
        service_get_all.return_value = services

        min_obj = objects.Service.get_minimum_obj_version(self.context)
        self.assertEqual('1.2', min_obj)
        service_get_all.assert_called_once_with(self.context, binary=None,
                                                disabled=None)

    @mock.patch('cinder.db.sqlalchemy.api.cluster_get')
    def test_lazy_loading_cluster_field(self, cluster_get):
        cluster_orm = fake_cluster.fake_cluster_orm(name='mycluster')
        cluster_get.return_value = cluster_orm
        cluster = objects.Cluster._from_db_object(self.context,
                                                  objects.Cluster(),
                                                  cluster_orm)

        service = fake_service.fake_service_obj(self.context,
                                                cluster_name='mycluster')
        self.assertEqual(cluster, service.cluster)
        cluster_get.assert_called_once_with(self.context, None,
                                            name='mycluster')

    def test_service_is_up(self):
        # NOTE(mdovgal): don't use @ddt.data with the real timestamp value
        # for this test.
        # When using ddt decorators ddt.data seems to have been calculated
        # not at the time of test's execution but at the tests's beginning.
        # And this one depends on utcnow func. So it won't be utcnow at the
        # execution moment and the result will be unexpected.
        down_time = 5
        self.flags(service_down_time=down_time)

        # test if service is up
        service = fake_service.fake_service_obj(self.context)
        self.assertTrue(service.is_up)

        service.updated_at = timeutils.utcnow()
        self.assertTrue(service.is_up)

        # test is service is down now
        past_time = timeutils.utcnow() - datetime.timedelta(seconds=64)
        service.updated_at = past_time
        self.assertFalse(service.is_up)


class TestServiceList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.db.service_get_all')
    def test_get_all(self, service_get_all):
        db_service = fake_service.fake_db_service()
        service_get_all.return_value = [db_service]

        filters = {'host': 'host', 'binary': 'foo', 'disabled': False}
        services = objects.ServiceList.get_all(self.context, filters)
        service_get_all.assert_called_once_with(self.context, **filters)
        self.assertEqual(1, len(services))
        TestService._compare(self, db_service, services[0])

    @mock.patch('cinder.db.service_get_all')
    def test_get_all_by_topic(self, service_get_all):
        db_service = fake_service.fake_db_service()
        service_get_all.return_value = [db_service]

        services = objects.ServiceList.get_all_by_topic(
            self.context, 'foo', 'bar')
        service_get_all.assert_called_once_with(
            self.context, topic='foo', disabled='bar')
        self.assertEqual(1, len(services))
        TestService._compare(self, db_service, services[0])

    @mock.patch('cinder.db.service_get_all')
    def test_get_all_by_binary(self, service_get_all):
        db_service = fake_service.fake_db_service()
        service_get_all.return_value = [db_service]

        services = objects.ServiceList.get_all_by_binary(
            self.context, 'foo', 'bar')
        service_get_all.assert_called_once_with(
            self.context, binary='foo', disabled='bar')
        self.assertEqual(1, len(services))
        TestService._compare(self, db_service, services[0])



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from cinder import exception
from cinder.tests.unit import fake_group
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import powerstore


class TestVolumeGroupCreateDeleteUpdate(powerstore.TestPowerStoreDriver):
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_chap_config")
    def setUp(self, mock_chap):
        super(TestVolumeGroupCreateDeleteUpdate, self).setUp()
        self.driver.check_for_setup_error()
        self.volume1 = fake_volume.fake_volume_obj(
            self.context,
            host="host@backend",
            provider_id="fake_id",
            size=8
        )
        self.volume2 = fake_volume.fake_volume_obj(
            self.context,
            host="host@backend",
            provider_id="fake_id",
            size=8
        )
        self.group = fake_group.fake_group_obj(
            self.context,
        )

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.create_vg")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_create_volume_group(self, mock_is_cg, mock_create):
        mock_create.return_value = "fake_id"
        mock_is_cg.return_value = True
        self.driver.create_group(self.context, self.group)

    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_create_volume_group_fallback_to_generic(self, mock_is_cg):
        mock_is_cg.return_value = False
        self.assertRaises(NotImplementedError,
                          self.driver.create_group,
                          self.context,
                          self.group)

    @mock.patch("requests.request")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_create_volume_group_bad_status(self,
                                            mock_is_cg,
                                            mock_create_request):
        mock_create_request.return_value = powerstore.MockResponse(rc=400)
        error = self.assertRaises(exception.VolumeBackendAPIException,
                                  self.driver.create_group,
                                  self.context,
                                  self.group)
        self.assertIn("Failed to create PowerStore volume group", error.msg)

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.delete_volume_or_snapshot")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_delete_volume_group(self, mock_is_cg, mock_get_id, mock_delete):
        self.driver.delete_group(self.context, self.group, [])

    @mock.patch("requests.request")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_delete_volume_group_bad_status(self,
                                            mock_is_cg,
                                            mock_get_id,
                                            mock_delete):
        mock_delete.return_value = powerstore.MockResponse(rc=400)
        error = self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.delete_group,
            self.context,
            self.group,
            []
        )
        self.assertIn("Failed to delete PowerStore volume group", error.msg)

    @mock.patch("cinder.objects.volume.Volume.is_replicated")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_update_volume_group_add_replicated_volumes(self,
                                                        mock_is_cg,
                                                        mock_replicated):
        mock_replicated.return_value = True
        self.assertRaises(exception.InvalidVolume,
                          self.driver.update_group,
                          self.context,
                          self.group,
                          [self.volume1],
                          [])

    @mock.patch("requests.request")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_update_volume_group_add_volumes_bad_status(self,
                                                        mock_is_cg,
                                                        mock_get_vg_id,
                                                        mock_add_volumes):
        mock_add_volumes.return_value = powerstore.MockResponse(rc=400)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.update_group,
                          self.context,
                          self.group,
                          [self.volume1],
                          [])

    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.remove_volumes_from_vg")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.add_volumes_to_vg")
    @mock.patch("cinder.volume.drivers.dell_emc.powerstore.client."
                "PowerStoreClient.get_vg_id_by_name")
    @mock.patch("cinder.volume.volume_utils.is_group_a_cg_snapshot_type")
    def test_update_volume_group_add_remove_volumes(self,
                                                    mock_is_cg,
                                                    mock_get_vg_id,
                                                    mock_add_volumes,
                                                    mock_remove_volumes):
        self.driver.update_group(self.context,
                                 self.group,
                                 add_volumes=[self.volume1],
                                 remove_volumes=[self.volume2])



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from cinder.objects.fields import VolumeAttachStatus
from cinder.scheduler import filters
from cinder.scheduler.filters import extra_specs_ops

LOG = logging.getLogger(__name__)


class CapabilitiesFilter(filters.BaseBackendFilter):
    """BackendFilter to work with resource (instance & volume) type records."""

    def _satisfies_extra_specs(self, capabilities, filter_properties):
        """Check if capabilities satisfy resource type requirements.

        Check that the capabilities provided by the services satisfy
        the extra specs associated with the resource type.
        """

        req_spec = filter_properties.get('request_spec')
        if req_spec and req_spec.get('operation') == 'extend_volume':
            # NOTE(erlon): By default, cinder considers that every backend
            # supports volume online extending. Those backends that don't
            # support it should report online_extend_support=False.
            online_extends = capabilities.get('online_extend_support', True)
            if online_extends is False:
                vol_prop = req_spec.get('volume_properties')
                attach_status = vol_prop.get('attach_status')
                if attach_status != VolumeAttachStatus.DETACHED:
                    LOG.debug("Backend doesn't support attached volume extend")
                    return False

        resource_type = filter_properties.get('resource_type')
        if not resource_type:
            return True

        extra_specs = resource_type.get('extra_specs', [])
        if not extra_specs:
            return True

        for key, req in extra_specs.items():

            # Either not scoped format, or in capabilities scope
            scope = key.split(':')

            # Ignore scoped (such as vendor-specific) capabilities
            if len(scope) > 1 and scope[0] != "capabilities":
                continue
            # Strip off prefix if spec started with 'capabilities:'
            elif scope[0] == "capabilities":
                del scope[0]

            cap = capabilities
            for index in range(len(scope)):
                try:
                    cap = cap[scope[index]]
                except (TypeError, KeyError):
                    LOG.debug("Backend doesn't provide capability '%(cap)s' ",
                              {'cap': scope[index]})
                    return False

            # Make all capability values a list so we can handle lists
            cap_list = [cap] if not isinstance(cap, list) else cap

            # Loop through capability values looking for any match
            for cap_value in cap_list:
                if extra_specs_ops.match(cap_value, req):
                    break
            else:
                # Nothing matched, so bail out
                LOG.debug('Volume type extra spec requirement '
                          '"%(key)s=%(req)s" does not match reported '
                          'capability "%(cap)s"',
                          {'key': key, 'req': req, 'cap': cap})
                return False
        return True

    def backend_passes(self, backend_state, filter_properties):
        """Return a list of backends that can create resource_type."""
        # Note(zhiteng) Currently only Cinder and Nova are using
        # this filter, so the resource type is either instance or
        # volume.
        if not self._satisfies_extra_specs(backend_state.capabilities,
                                           filter_properties):
            LOG.debug("%(backend_state)s fails resource_type extra_specs "
                      "requirements", {'backend_state': backend_state})
            return False
        return True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from cinder import exception
from cinder.i18n import _


class InvalidBackend(exception.VolumeDriverException):
    message = _("Backend doesn't exist (%(backend)s)")


class ConnectionError(exception.VolumeDriverException):
    message = "%(message)s"


class AuthenticationError(exception.VolumeDriverException):
    message = "%(message)s"


class NotEnoughSpace(exception.VolumeDriverException):
    message = _("Not enough space on backend (%(backend)s)")


class RequestError(exception.VolumeDriverException):
    message = "%(message)s"


class NotTargetPortal(exception.VolumeDriverException):
    message = _("No active iSCSI portals with supplied iSCSI IPs")



# Copyright 2016 EMC Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volumes snapshots V3 API."""

import ast
from http import HTTPStatus

from oslo_log import log as logging
from oslo_utils import strutils
from webob import exc

from cinder.api import api_utils
from cinder.api import common
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import snapshots as snapshot
from cinder.api.v2 import snapshots as snapshots_v2
from cinder.api.v3.views import snapshots as snapshot_views
from cinder.api import validation
from cinder import utils

LOG = logging.getLogger(__name__)

SNAPSHOT_IN_USE_FLAG_MSG = (
    f"Since microversion {mv.SNAPSHOT_IN_USE} the 'force' flag is "
    "invalid for this request.  For backward compatability, however, when "
    "the 'force' flag is passed with a value evaluating to True, it is "
    "silently ignored.")


class SnapshotsController(snapshots_v2.SnapshotsController):
    """The Snapshots API controller for the OpenStack API."""

    _view_builder_class = snapshot_views.ViewBuilder

    def _get_snapshot_filter_options(self):
        """returns tuple of valid filter options"""

        return 'status', 'volume_id', 'name', 'metadata'

    def _format_snapshot_filter_options(self, search_opts):
        """Convert valid filter options to correct expected format"""

        # Get the dict object out of queried metadata
        # convert metadata query value from string to dict
        if 'metadata' in search_opts.keys():
            try:
                search_opts['metadata'] = ast.literal_eval(
                    search_opts['metadata'])
            except (ValueError, SyntaxError):
                LOG.debug('Could not evaluate value %s, assuming string',
                          search_opts['metadata'])

        if 'use_quota' in search_opts:
            search_opts['use_quota'] = utils.get_bool_param('use_quota',
                                                            search_opts)

    MV_ADDED_FILTERS = (
        (mv.get_prior_version(mv.SNAPSHOT_LIST_METADATA_FILTER), 'metadata'),
        # REST API receives consumes_quota, but process_general_filtering
        # transforms it into use_quota
        (mv.get_prior_version(mv.USE_QUOTA), 'use_quota'),
    )

    @common.process_general_filtering('snapshot')
    def _process_snapshot_filtering(self, context=None, filters=None,
                                    req_version=None):
        """Formats allowed filters"""
        for version, field in self.MV_ADDED_FILTERS:
            if req_version.matches(None, version):
                filters.pop(field, None)

        # Filter out invalid options
        allowed_search_options = self._get_snapshot_filter_options()

        api_utils.remove_invalid_filter_options(context, filters,
                                                allowed_search_options)

    def _items(self, req, is_detail=True):
        """Returns a list of snapshots, transformed through view builder."""
        context = req.environ['cinder.context']
        # Pop out non search_opts and create local variables
        search_opts = req.GET.copy()
        sort_keys, sort_dirs = common.get_sort_params(search_opts)
        marker, limit, offset = common.get_pagination_params(search_opts)

        req_version = req.api_version_request
        show_count = False
        if req_version.matches(
                mv.SUPPORT_COUNT_INFO) and 'with_count' in search_opts:
            show_count = utils.get_bool_param('with_count', search_opts)
            search_opts.pop('with_count')

        # process filters
        self._process_snapshot_filtering(context=context,
                                         filters=search_opts,
                                         req_version=req_version)
        # process snapshot filters to appropriate formats if required
        self._format_snapshot_filter_options(search_opts)

        req_version = req.api_version_request
        if req_version.matches(mv.SNAPSHOT_SORT, None) and 'name' in sort_keys:
            sort_keys[sort_keys.index('name')] = 'display_name'

        # NOTE(thingee): v3 API allows name instead of display_name
        if 'name' in search_opts:
            search_opts['display_name'] = search_opts.pop('name')

        snapshots = self.volume_api.get_all_snapshots(
            context,
            search_opts=search_opts.copy(),
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            offset=offset)
        total_count = None
        if show_count:
            total_count = self.volume_api.calculate_resource_count(
                context, 'snapshot', search_opts)

        req.cache_db_snapshots(snapshots.objects)

        if is_detail:
            snapshots = self._view_builder.detail_list(req, snapshots.objects,
                                                       total_count)
        else:
            snapshots = self._view_builder.summary_list(req, snapshots.objects,
                                                        total_count)
        return snapshots

    @wsgi.response(HTTPStatus.ACCEPTED)
    @validation.schema(snapshot.create)
    def create(self, req, body):
        """Creates a new snapshot."""
        kwargs = {}
        context = req.environ['cinder.context']
        snapshot = body['snapshot']
        kwargs['metadata'] = snapshot.get('metadata', None)
        volume_id = snapshot['volume_id']
        volume = self.volume_api.get(context, volume_id)
        req_version = req.api_version_request
        force_flag = snapshot.get('force')
        force = False
        if force_flag is not None:
            # note: this won't raise because it passed schema validation
            force = strutils.bool_from_string(force_flag, strict=True)

            if req_version.matches(mv.SNAPSHOT_IN_USE):
                # strictly speaking, the 'force' flag is invalid for
                # mv.SNAPSHOT_IN_USE, but we silently ignore a True
                # value for backward compatibility
                if force is False:
                    raise exc.HTTPBadRequest(
                        explanation=SNAPSHOT_IN_USE_FLAG_MSG)

        LOG.info("Create snapshot from volume %s", volume_id)

        self.validate_name_and_description(snapshot, check_length=False)
        if 'name' in snapshot:
            snapshot['display_name'] = snapshot.pop('name')

        if force:
            new_snapshot = self.volume_api.create_snapshot_force(
                context,
                volume,
                snapshot.get('display_name'),
                snapshot.get('description'),
                **kwargs)
        else:
            if req_version.matches(mv.SNAPSHOT_IN_USE):
                kwargs['allow_in_use'] = True

            new_snapshot = self.volume_api.create_snapshot(
                context,
                volume,
                snapshot.get('display_name'),
                snapshot.get('description'),
                **kwargs)
        req.cache_db_snapshot(new_snapshot)

        return self._view_builder.detail(req, new_snapshot)


def create_resource(ext_mgr):
    return wsgi.Resource(SnapshotsController(ext_mgr))



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Schema for V3 volume type encryption API."""

import copy

from cinder.api.validation import parameter_types

create = {
    'type': 'object',
    'properties': {
        'encryption': {
            'type': 'object',
            'properties': {
                'key_size': parameter_types.key_size,
                'provider': {'type': 'string', 'minLength': 0,
                             'maxLength': 255},
                'control_location': {'enum': ['front-end', 'back-end']},
                'cipher': {'type': ['string', 'null'],
                           'minLength': 0, 'maxLength': 255},
            },
            'required': ['provider', 'control_location'],
            'additionalProperties': True,
        },
    },
    'required': ['encryption'],
    'additionalProperties': False,
}


update = copy.deepcopy(create)
update['properties']['encryption']['required'] = []



# Copyright (C) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for db purge."""

import datetime
import uuid

from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_utils import timeutils
from sqlalchemy.dialects import sqlite

from cinder import context
from cinder import db
from cinder.db.sqlalchemy import api as db_api
from cinder import exception
from cinder.tests.unit import test


class PurgeDeletedTest(test.TestCase):

    def setUp(self):
        super(PurgeDeletedTest, self).setUp()
        self.context = context.get_admin_context()
        self.engine = db_api.get_engine()
        self.session = db_api.get_session()
        self.conn = self.engine.connect()
        self.volumes = sqlalchemyutils.get_table(
            self.engine, "volumes")
        # The volume_metadata table has a FK of volume_id
        self.vm = sqlalchemyutils.get_table(
            self.engine, "volume_metadata")

        self.vol_types = sqlalchemyutils.get_table(
            self.engine, "volume_types")
        # The volume_type_projects table has a FK of volume_type_id
        self.vol_type_proj = sqlalchemyutils.get_table(
            self.engine, "volume_type_projects")

        self.snapshots = sqlalchemyutils.get_table(
            self.engine, "snapshots")

        self.sm = sqlalchemyutils.get_table(
            self.engine, "snapshot_metadata")

        self.vgm = sqlalchemyutils.get_table(
            self.engine, "volume_glance_metadata")

        self.qos = sqlalchemyutils.get_table(
            self.engine, "quality_of_service_specs")

        self.uuidstrs = []
        for unused in range(6):
            self.uuidstrs.append(uuid.uuid4().hex)
        # Add 6 rows to table
        for uuidstr in self.uuidstrs:
            ins_stmt = self.volumes.insert().values(id=uuidstr,
                                                    volume_type_id=uuidstr)
            self.conn.execute(ins_stmt)
            ins_stmt = self.vm.insert().values(volume_id=uuidstr)
            self.conn.execute(ins_stmt)
            ins_stmt = self.vgm.insert().values(
                volume_id=uuidstr, key='image_name', value='test')
            self.conn.execute(ins_stmt)

            ins_stmt = self.vol_types.insert().values(id=uuidstr)
            self.conn.execute(ins_stmt)
            ins_stmt = self.vol_type_proj.insert().\
                values(volume_type_id=uuidstr)
            self.conn.execute(ins_stmt)

            ins_stmt = self.snapshots.insert().values(
                id=uuidstr, volume_id=uuidstr,
                volume_type_id=uuidstr)
            self.conn.execute(ins_stmt)
            ins_stmt = self.sm.insert().values(snapshot_id=uuidstr)
            self.conn.execute(ins_stmt)

            ins_stmt = self.vgm.insert().values(
                snapshot_id=uuidstr, key='image_name', value='test')
            self.conn.execute(ins_stmt)

            ins_stmt = self.qos.insert().values(
                id=uuidstr, key='QoS_Specs_Name', value='test')
            self.conn.execute(ins_stmt)

            ins_stmt = self.vol_types.insert().values(
                id=uuid.uuid4().hex, qos_specs_id=uuidstr)
            self.conn.execute(ins_stmt)

            ins_stmt = self.qos.insert().values(
                id=uuid.uuid4().hex, specs_id=uuidstr, key='desc',
                value='test')
            self.conn.execute(ins_stmt)

        # Set 5 of them deleted
        # 2 are 60 days ago, 2 are 20 days ago, one is just now.
        now = timeutils.utcnow()
        old = timeutils.utcnow() - datetime.timedelta(days=20)
        older = timeutils.utcnow() - datetime.timedelta(days=60)

        make_vol_now = self.volumes.update().\
            where(self.volumes.c.id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_vol_old = self.volumes.update().\
            where(self.volumes.c.id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_vol_older = self.volumes.update().\
            where(self.volumes.c.id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)
        make_vol_meta_now = self.vm.update().\
            where(self.vm.c.volume_id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_vol_meta_old = self.vm.update().\
            where(self.vm.c.volume_id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_vol_meta_older = self.vm.update().\
            where(self.vm.c.volume_id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)

        make_vol_types_now = self.vol_types.update().\
            where(self.vol_types.c.id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_vol_types_old = self.vol_types.update().\
            where(self.vol_types.c.id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_vol_types_older = self.vol_types.update().\
            where(self.vol_types.c.id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)
        make_vol_type_proj_now = self.vol_type_proj.update().\
            where(self.vol_type_proj.c.volume_type_id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_vol_type_proj_old = self.vol_type_proj.update().\
            where(self.vol_type_proj.c.volume_type_id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_vol_type_proj_older = self.vol_type_proj.update().\
            where(self.vol_type_proj.c.volume_type_id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)

        make_snap_now = self.snapshots.update().\
            where(self.snapshots.c.id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_snap_old = self.snapshots.update().\
            where(self.snapshots.c.id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_snap_older = self.snapshots.update().\
            where(self.snapshots.c.id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)

        make_snap_meta_now = self.sm.update().\
            where(self.sm.c.snapshot_id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_snap_meta_old = self.sm.update().\
            where(self.sm.c.snapshot_id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_snap_meta_older = self.sm.update().\
            where(self.sm.c.snapshot_id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)

        make_vol_glance_meta_now = self.vgm.update().\
            where(self.vgm.c.volume_id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_vol_glance_meta_old = self.vgm.update().\
            where(self.vgm.c.volume_id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_vol_glance_meta_older = self.vgm.update().\
            where(self.vgm.c.volume_id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)
        make_snap_glance_meta_now = self.vgm.update().\
            where(self.vgm.c.snapshot_id.in_(self.uuidstrs[0:1]))\
            .values(deleted_at=now, deleted=True)
        make_snap_glance_meta_old = self.vgm.update().\
            where(self.vgm.c.snapshot_id.in_(self.uuidstrs[1:3]))\
            .values(deleted_at=old, deleted=True)
        make_snap_glance_meta_older = self.vgm.update().\
            where(self.vgm.c.snapshot_id.in_(self.uuidstrs[4:6]))\
            .values(deleted_at=older, deleted=True)

        make_qos_now = self.qos.update().where(
            self.qos.c.id.in_(self.uuidstrs[0:1])).values(deleted_at=now,
                                                          deleted=True)
        make_qos_old = self.qos.update().where(
            self.qos.c.id.in_(self.uuidstrs[1:3])).values(deleted_at=old,
                                                          deleted=True)
        make_qos_older = self.qos.update().where(
            self.qos.c.id.in_(self.uuidstrs[4:6])).values(deleted_at=older,
                                                          deleted=True)

        make_qos_child_record_now = self.qos.update().where(
            self.qos.c.specs_id.in_(self.uuidstrs[0:1])).values(
            deleted_at=now, deleted=True)
        make_qos_child_record_old = self.qos.update().where(
            self.qos.c.specs_id.in_(self.uuidstrs[1:3])).values(
            deleted_at=old, deleted=True)
        make_qos_child_record_older = self.qos.update().where(
            self.qos.c.specs_id.in_(self.uuidstrs[4:6])).values(
            deleted_at=older, deleted=True)

        make_vol_types1_now = self.vol_types.update().where(
            self.vol_types.c.qos_specs_id.in_(self.uuidstrs[0:1])).values(
            deleted_at=now, deleted=True)
        make_vol_types1_old = self.vol_types.update().where(
            self.vol_types.c.qos_specs_id.in_(self.uuidstrs[1:3])).values(
            deleted_at=old, deleted=True)
        make_vol_types1_older = self.vol_types.update().where(
            self.vol_types.c.qos_specs_id.in_(self.uuidstrs[4:6])).values(
            deleted_at=older, deleted=True)

        self.conn.execute(make_vol_now)
        self.conn.execute(make_vol_old)
        self.conn.execute(make_vol_older)
        self.conn.execute(make_vol_meta_now)
        self.conn.execute(make_vol_meta_old)
        self.conn.execute(make_vol_meta_older)

        self.conn.execute(make_vol_types_now)
        self.conn.execute(make_vol_types_old)
        self.conn.execute(make_vol_types_older)
        self.conn.execute(make_vol_type_proj_now)
        self.conn.execute(make_vol_type_proj_old)
        self.conn.execute(make_vol_type_proj_older)

        self.conn.execute(make_snap_now)
        self.conn.execute(make_snap_old)
        self.conn.execute(make_snap_older)
        self.conn.execute(make_snap_meta_now)
        self.conn.execute(make_snap_meta_old)
        self.conn.execute(make_snap_meta_older)

        self.conn.execute(make_vol_glance_meta_now)
        self.conn.execute(make_vol_glance_meta_old)
        self.conn.execute(make_vol_glance_meta_older)
        self.conn.execute(make_snap_glance_meta_now)
        self.conn.execute(make_snap_glance_meta_old)
        self.conn.execute(make_snap_glance_meta_older)

        self.conn.execute(make_qos_now)
        self.conn.execute(make_qos_old)
        self.conn.execute(make_qos_older)

        self.conn.execute(make_qos_child_record_now)
        self.conn.execute(make_qos_child_record_old)
        self.conn.execute(make_qos_child_record_older)

        self.conn.execute(make_vol_types1_now)
        self.conn.execute(make_vol_types1_old)
        self.conn.execute(make_vol_types1_older)

    def test_purge_deleted_rows_in_zero_age_in(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            self.conn.execute("PRAGMA foreign_keys = ON")
        # Purge at age_in_days=0, should delete one more row
        db.purge_deleted_rows(self.context, age_in_days=0)

        vol_rows = self.session.query(self.volumes).count()
        vol_meta_rows = self.session.query(self.vm).count()
        vol_type_rows = self.session.query(self.vol_types).count()
        vol_type_proj_rows = self.session.query(self.vol_type_proj).count()
        snap_rows = self.session.query(self.snapshots).count()
        snap_meta_rows = self.session.query(self.sm).count()
        vol_glance_meta_rows = self.session.query(self.vgm).count()
        qos_rows = self.session.query(self.qos).count()

        # Verify that we only have 1 rows now
        self.assertEqual(1, vol_rows)
        self.assertEqual(1, vol_meta_rows)
        self.assertEqual(3, vol_type_rows)
        self.assertEqual(1, vol_type_proj_rows)
        self.assertEqual(1, snap_rows)
        self.assertEqual(1, snap_meta_rows)
        self.assertEqual(2, vol_glance_meta_rows)
        self.assertEqual(2, qos_rows)

    def test_purge_deleted_rows_old(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            self.conn.execute("PRAGMA foreign_keys = ON")
        # Purge at 30 days old, should only delete 2 rows
        db.purge_deleted_rows(self.context, age_in_days=30)

        vol_rows = self.session.query(self.volumes).count()
        vol_meta_rows = self.session.query(self.vm).count()
        vol_type_rows = self.session.query(self.vol_types).count()
        vol_type_proj_rows = self.session.query(self.vol_type_proj).count()
        snap_rows = self.session.query(self.snapshots).count()
        snap_meta_rows = self.session.query(self.sm).count()
        vol_glance_meta_rows = self.session.query(self.vgm).count()
        qos_rows = self.session.query(self.qos).count()

        # Verify that we only deleted 2
        self.assertEqual(4, vol_rows)
        self.assertEqual(4, vol_meta_rows)
        self.assertEqual(9, vol_type_rows)
        self.assertEqual(4, vol_type_proj_rows)
        self.assertEqual(4, snap_rows)
        self.assertEqual(4, snap_meta_rows)
        self.assertEqual(8, vol_glance_meta_rows)
        self.assertEqual(8, qos_rows)

    def test_purge_deleted_rows_older(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            self.conn.execute("PRAGMA foreign_keys = ON")
        # Purge at 10 days old now, should delete 2 more rows
        db.purge_deleted_rows(self.context, age_in_days=10)

        vol_rows = self.session.query(self.volumes).count()
        vol_meta_rows = self.session.query(self.vm).count()
        vol_type_rows = self.session.query(self.vol_types).count()
        vol_type_proj_rows = self.session.query(self.vol_type_proj).count()
        snap_rows = self.session.query(self.snapshots).count()
        snap_meta_rows = self.session.query(self.sm).count()
        vol_glance_meta_rows = self.session.query(self.vgm).count()
        qos_rows = self.session.query(self.qos).count()

        # Verify that we only have 2 rows now
        self.assertEqual(2, vol_rows)
        self.assertEqual(2, vol_meta_rows)
        self.assertEqual(5, vol_type_rows)
        self.assertEqual(2, vol_type_proj_rows)
        self.assertEqual(2, snap_rows)
        self.assertEqual(2, snap_meta_rows)
        self.assertEqual(4, vol_glance_meta_rows)
        self.assertEqual(4, qos_rows)

    def test_purge_deleted_rows_bad_args(self):
        # Test with no age argument
        self.assertRaises(TypeError, db.purge_deleted_rows, self.context)
        # Test purge with non-integer
        self.assertRaises(exception.InvalidParameterValue,
                          db.purge_deleted_rows, self.context,
                          age_in_days='ten')

    def test_purge_deleted_rows_integrity_failure(self):
        dialect = self.engine.url.get_dialect()
        if dialect == sqlite.dialect:
            self.conn.execute("PRAGMA foreign_keys = ON")

        # add new entry in volume and volume_admin_metadata for
        # integrity check
        uuid_str = uuid.uuid4().hex
        ins_stmt = self.volumes.insert().values(id=uuid_str,
                                                volume_type_id=uuid_str)
        self.conn.execute(ins_stmt)
        ins_stmt = self.vm.insert().values(volume_id=uuid_str)
        self.conn.execute(ins_stmt)

        # set volume record to deleted 20 days ago
        old = timeutils.utcnow() - datetime.timedelta(days=20)
        make_old = self.volumes.update().where(
            self.volumes.c.id.in_([uuid_str])).values(deleted_at=old,
                                                      deleted=True)
        self.conn.execute(make_old)

        # Verify that purge_deleted_rows fails due to Foreign Key constraint
        self.assertRaises(db_exc.DBReferenceError, db.purge_deleted_rows,
                          self.context, age_in_days=10)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock

import ddt

from cinder.api.contrib import admin_actions
from cinder.api.contrib import volume_actions
from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.v3 import volumes
from cinder import exception
from cinder.objects import fields
from cinder.policies import volume_actions as policy
from cinder.policies import volumes as volume_policy
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants
from cinder.tests.unit.policies import base
from cinder.tests.unit.policies import test_base
from cinder.tests.unit import utils as test_utils
from cinder.volume import api as volume_api
from cinder.volume import manager as volume_manager


@ddt.ddt
class VolumeActionsPolicyTest(base.BasePolicyTest):
    authorized_users = [
        'legacy_admin',
        'legacy_owner',
        'system_admin',
        'project_admin',
        'project_member',
        'project_reader',
        'project_foo',
    ]
    unauthorized_users = [
        'system_member',
        'system_reader',
        'system_foo',
        'other_project_member',
        'other_project_reader',
    ]

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]

    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)

        self.ext_mgr = extensions.ExtensionManager()
        self.controller = volume_actions.VolumeActionsController(self.ext_mgr)
        self.admin_controller = admin_actions.VolumeAdminController(
            self.ext_mgr)
        self.volume_controller = volumes.VolumeController(self.ext_mgr)
        self.manager = volume_manager.VolumeManager()
        self.manager.driver = mock.MagicMock()
        self.manager.driver.initialize_connection = mock.MagicMock()
        self.manager.driver.initialize_connection.side_effect = (
            self._initialize_connection)
        self.api_path = '/v3/%s/volumes' % (self.project_id)
        self.api_version = mv.BASE_VERSION

    def _initialize_connection(self, volume, connector):
        return {'data': connector}

    def _create_volume(self, attached=False, **kwargs):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)
        volume = test_utils.create_volume(self.project_member_context,
                                          volume_type_id=vol_type.id,
                                          testcase_instance=self, **kwargs)

        if attached:
            volume = test_utils.attach_volume(self.project_member_context,
                                              volume.id,
                                              fake_constants.INSTANCE_ID,
                                              'fake_host',
                                              'fake_mountpoint')
        return volume

    @ddt.data(*base.all_users)
    def test_extend_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.EXTEND_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-extend": {
                "new_size": 3
            }
        }

        # DB validations will throw VolumeNotFound for some contexts
        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._extend, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_extend_attached_policy(self, user_id):
        volume = self._create_volume(attached=True)
        rule_name = policy.EXTEND_ATTACHED_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.VOLUME_EXTEND_INUSE)
        req.method = 'POST'
        body = {
            "os-extend": {
                "new_size": 3
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._extend, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_revert_policy(self, user_id):
        volume = self._create_volume()
        snap = test_utils.create_snapshot(
            self.project_member_context,
            volume.id,
            status=fields.SnapshotStatus.AVAILABLE,
            testcase_instance=self)
        rule_name = policy.REVERT_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.VOLUME_REVERT)
        req.method = 'POST'
        body = {
            "revert": {
                "snapshot_id": snap.id
            }
        }

        # Relax the volume:GET_POLICY in order to get past that check.
        self.policy.set_rules({volume_policy.GET_POLICY: ""},
                              overwrite=False)

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.volume_controller.revert, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_reset_policy(self, user_id):
        volume = self._create_volume(attached=True)
        rule_name = policy.RESET_STATUS
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-reset_status": {
                "status": "available",
                "attach_status": "detached",
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._reset_status, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_retype_policy(self, user_id):
        volume = self._create_volume()
        test_utils.create_volume_type(self.project_admin_context,
                                      name='another_vol_type',
                                      testcase_instance=self)
        rule_name = policy.RETYPE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-retype": {
                "new_type": "another_vol_type",
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._retype, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_update_readonly_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.UPDATE_READONLY_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-update_readonly_flag": {
                "readonly": True
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller._volume_readonly_update, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_force_delete_policy(self, user_id):
        volume = self._create_volume()
        rule_name = policy.FORCE_DELETE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-force_delete": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._force_delete, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.detach_volume')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.terminate_connection')
    def test_force_detach_policy(self, user_id,
                                 mock_terminate_connection,
                                 mock_detach_volume):
        # Redirect the RPC calls directly to the volume manager.
        # The volume manager needs the volume.id, not the volume.
        def detach_volume(ctxt, volume, connector, force=False):
            return self.manager.detach_volume(ctxt, volume.id,
                                              attachment_id=None,
                                              volume=None)

        def terminate_connection(ctxt, volume, connector, force=False):
            return self.manager.terminate_connection(ctxt, volume.id,
                                                     connector, force)

        mock_detach_volume.side_effect = detach_volume
        mock_terminate_connection.side_effect = terminate_connection

        volume = self._create_volume(attached=True)
        rule_name = policy.FORCE_DETACH_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-force_detach": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._force_detach, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.copy_volume_to_image')
    @mock.patch('cinder.image.glance.GlanceImageService.create')
    def test_upload_image_policy(self, user_id,
                                 mock_image_create,
                                 mock_copy_volume_to_image):
        # Redirect the RPC calls directly to the volume manager.
        # The volume manager needs the volume.id, not the volume.
        def copy_volume_to_image(ctxt, volume, image_meta):
            return self.manager.copy_volume_to_image(ctxt, volume.id,
                                                     image_meta)

        mock_copy_volume_to_image.side_effect = copy_volume_to_image

        volume = self._create_volume(status='available')
        rule_name = policy.UPLOAD_IMAGE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-volume_upload_image": {
                "image_name": "test",
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller._volume_upload_image, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.copy_volume_to_image')
    @mock.patch('cinder.image.glance.GlanceImageService.create')
    def test_upload_public_policy(self, user_id,
                                  mock_image_create,
                                  mock_copy_volume_to_image):
        # Redirect the RPC calls directly to the volume manager.
        # The volume manager needs the volume.id, not the volume.
        def copy_volume_to_image(ctxt, volume, image_meta):
            return self.manager.copy_volume_to_image(ctxt, volume.id,
                                                     image_meta)

        mock_copy_volume_to_image.side_effect = copy_volume_to_image

        volume = self._create_volume(status='available')
        rule_name = policy.UPLOAD_PUBLIC_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=mv.UPLOAD_IMAGE_PARAMS)
        req.method = 'POST'
        body = {
            "os-volume_upload_image": {
                "image_name": "test",
                "visibility": "public",
            }
        }

        # Relax the UPLOAD_IMAGE_POLICY in order to get past that check.
        self.policy.set_rules({policy.UPLOAD_IMAGE_POLICY: ""},
                              overwrite=False)

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller._volume_upload_image, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.objects.Service.get_by_id')
    def test_migrate_policy(self, user_id, mock_get_service_by_id):
        volume = self._create_volume()
        rule_name = policy.MIGRATE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-migrate_volume": {
                "host": "node1@lvm"
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.admin_controller._migrate_volume, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_migrate_complete_policy(self, user_id):
        volume = self._create_volume()
        # Can't use self._create_volume() because it would fail when
        # trying to create the volume type a second time.
        new_volume = test_utils.create_volume(self.project_member_context,
                                              testcase_instance=self)
        rule_name = policy.MIGRATE_COMPLETE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-migrate_volume_completion": {
                "new_volume": new_volume.id
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(
            user_id, self.authorized_admins, self.unauthorized_admins,
            unauthorized_exceptions, rule_name,
            self.admin_controller._migrate_volume_completion, req,
            id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.attach_volume')
    def test_attach_policy(self, user_id, mock_attach_volume):
        def attach_volume(context, volume, instance_uuid, host_name,
                          mountpoint, mode):
            return self.manager.attach_volume(context, volume.id,
                                              instance_uuid, host_name,
                                              mountpoint, mode)

        mock_attach_volume.side_effect = attach_volume

        volume = self._create_volume(status='available')
        rule_name = policy.ATTACH_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-attach": {
                "instance_uuid": fake_constants.INSTANCE_ID,
                "mountpoint": "/dev/vdc"
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._attach, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.detach_volume')
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.terminate_connection')
    def test_detach_policy(self, user_id,
                           mock_terminate_connection,
                           mock_detach_volume):
        # Redirect the RPC calls directly to the volume manager.
        # The volume manager needs the volume.id, not the volume.
        def detach_volume(ctxt, volume, connector, force=False):
            return self.manager.detach_volume(ctxt, volume.id,
                                              attachment_id=None,
                                              volume=None)

        def terminate_connection(ctxt, volume, connector, force=False):
            return self.manager.terminate_connection(ctxt, volume.id,
                                                     connector, force)

        mock_detach_volume.side_effect = detach_volume
        mock_terminate_connection.side_effect = terminate_connection

        volume = self._create_volume(attached=True)
        rule_name = policy.DETACH_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-detach": {
                "attachment_id": volume.volume_attachment[0].id
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._detach, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_begin_detaching_policy(self, user_id):
        volume = self._create_volume(status='in-use', attach_status='attached')
        rule_name = policy.BEGIN_DETACHING_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-begin_detaching": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._begin_detaching,
                                 req, id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_reserve_policy(self, user_id):
        volume = self._create_volume(status='available')
        rule_name = policy.RESERVE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-reserve": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._reserve, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_unreserve_policy(self, user_id):
        volume = self._create_volume(status='reserved')
        rule_name = policy.UNRESERVE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-unreserve": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._unreserve, req,
                                 id=volume.id, body=body)

    @ddt.data(*base.all_users)
    def test_roll_detaching_policy(self, user_id):
        volume = self._create_volume(status='detaching')
        rule_name = policy.ROLL_DETACHING_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-roll_detaching": {}
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name, self.controller._roll_detaching,
                                 req, id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.initialize_connection')
    def test_initialize_policy(self, user_id, mock_initialize_connection):
        def initialize_connection(*args):
            return self.manager.initialize_connection(*args)

        mock_initialize_connection.side_effect = initialize_connection

        volume = self._create_volume()
        rule_name = policy.INITIALIZE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-initialize_connection": {
                "connector": {
                    "platform": "x86_64",
                    "host": "node2",
                    "do_local_attach": False,
                    "ip": "192.168.13.101",
                    "os_type": "linux2",
                    "multipath": False,
                    "initiator": "iqn.1994-05.com.redhat:d16cbb5d31e5"
                }
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller._initialize_connection,
                                 req, id=volume.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch('cinder.volume.rpcapi.VolumeAPI.terminate_connection')
    def test_terminate_policy(self, user_id, mock_terminate_connection):
        def terminate_connection(ctxt, volume, connector, force=False):
            return self.manager.terminate_connection(ctxt, volume.id,
                                                     connector, force=False)

        mock_terminate_connection.side_effect = terminate_connection

        volume = self._create_volume()
        rule_name = policy.TERMINATE_POLICY
        url = '%s/%s/action' % (self.api_path, volume.id)
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {
            "os-terminate_connection": {
                "connector": {
                    "platform": "x86_64",
                    "host": "node2",
                    "do_local_attach": False,
                    "ip": "192.168.13.101",
                    "os_type": "linux2",
                    "multipath": False,
                    "initiator": "iqn.1994-05.com.redhat:d16cbb5d31e5"
                }
            }
        }

        unauthorized_exceptions = [
            exception.VolumeNotFound,
        ]

        self.common_policy_check(user_id, self.authorized_users,
                                 self.unauthorized_users,
                                 unauthorized_exceptions,
                                 rule_name,
                                 self.controller._terminate_connection,
                                 req, id=volume.id, body=body)


class VolumeActionsPolicySecureRbacTest(VolumeActionsPolicyTest):
    authorized_users = [
        'legacy_admin',
        'system_admin',
        'project_admin',
        'project_member',
    ]
    unauthorized_users = [
        'legacy_owner',
        'system_member',
        'system_foo',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)


class VolumeProtectionTests(test_base.CinderPolicyTests):
    def test_admin_can_extend_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_extend_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_extend_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_extend_attached_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_extend_attached_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_extend_attached_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-extend": {"new_size": "2"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body,
            microversion=mv.VOLUME_EXTEND_INUSE)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_retype_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        vol_type = self._create_fake_type(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_retype_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        vol_type = self._create_fake_type(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_retype_volume_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        vol_type = self._create_fake_type(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-retype": {"new_type": "%s" % vol_type.name,
                              "migration_policy": "never"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_update_readonly(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(
            admin_context, admin_metadata={"readonly": "False"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            admin_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_update_readonly(self):
        user_context = self.user_context

        volume = self._create_fake_volume(
            user_context, admin_metadata={"readonly": "False"})

        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            user_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_update_readonly_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(
            user_context, admin_metadata={"readonly": "False"})
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-update_readonly_flag": {"readonly": "True"}}
        response = self._get_request_response(
            non_owner_context, path, 'POST', body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_admin_can_force_delete_volumes(self, mock_volume):
        # Make sure administrators are authorized to force delete volumes
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }
        body = {"os-force_delete": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get_volume')
    def test_nonadmin_cannot_force_delete_volumes(self, mock_volume):
        # Make sure volumes only can be force deleted by admin
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }
        body = {"os-force_delete": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)

        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    def test_admin_can_attach_detach_volume(self, mock_detach, mock_attach):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    def test_owner_can_attach_detach_volume(self, mock_detach, mock_attach):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'attach_volume')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI, 'detach_volume')
    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_attach_detach_volume_for_others(self, mock_volume,
                                                          mock_detach,
                                                          mock_attach):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_reserve_unreserve_volume(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-reserve": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-unreserve": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_reserve_unreserve_volume(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-reserve": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-unreserve": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_reserve_unreserve_volume_for_others(self,
                                                              mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-attach": {"instance_uuid": fake_constants.UUID1,
                              "mountpoint": "/dev/vdc"}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

        body = {"os-detach": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    def test_admin_can_initialize_terminate_conn(self, mock_t, mock_i):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    def test_owner_can_initialize_terminate_conn(self, mock_t, mock_i):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context)
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.OK, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'initialize_connection')
    @mock.patch.object(volume_api.volume_rpcapi.VolumeAPI,
                       'terminate_connection')
    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_initialize_terminate_conn_for_others(self,
                                                               mock_volume,
                                                               mock_t,
                                                               mock_i):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context)
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-initialize_connection": {'connector': {}}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

        body = {"os-terminate_connection": {'connector': {}}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

    def test_admin_can_begin_roll_detaching(self):
        admin_context = self.admin_context

        volume = self._create_fake_volume(admin_context, status='in-use',
                                          attach_status='attached')
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': admin_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(admin_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    def test_owner_can_begin_roll_detaching(self):
        user_context = self.user_context

        volume = self._create_fake_volume(user_context, status='in-use',
                                          attach_status='attached')
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': user_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(user_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.ACCEPTED, response.status_int)

    @mock.patch.object(volume_api.API, 'get')
    def test_owner_cannot_begin_roll_detaching_for_others(self, mock_volume):
        user_context = self.user_context
        non_owner_context = self.other_user_context

        volume = self._create_fake_volume(user_context, status='in-use',
                                          attach_status='attached')
        mock_volume.return_value = volume
        path = '/v3/%(project_id)s/volumes/%(volume_id)s/action' % {
            'project_id': non_owner_context.project_id, 'volume_id': volume.id
        }

        body = {"os-begin_detaching": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)

        body = {"os-roll_detaching": {}}
        response = self._get_request_response(non_owner_context, path, 'POST',
                                              body=body)
        self.assertEqual(HTTPStatus.FORBIDDEN, response.status_int)



# Copyright (c) 2018 Hedvig, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Config(object):
    ReplicationPolicy = {
        0: "Agnostic",
        1: "RackAware",
        2: "DataCenterAware",
    }

    DiskResidence = {
        0: "Flash",
        1: "HDD",
    }

    # Default Port Configuration
    defaultHControllerPort_ = 50000

    # Default Cinder Configuration
    defaultCinderReplicationFactor = 3
    defaultCinderDedupEnable = False
    defaultCinderCompressEnable = False
    defaultCinderCacheEnable = False
    defaultCinderDiskResidence = DiskResidence[1]
    defaultCinderReplicationPolicy = ReplicationPolicy[0]
    retryCount = 5



#    Copyright 2015 Yahoo Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.objects import fields as c_fields


@base.CinderObjectRegistry.register
class ConsistencyGroup(base.CinderPersistentObject, base.CinderObject,
                       base.CinderObjectDictCompat, base.ClusteredObject):
    # Version 1.0: Initial version
    # Version 1.1: Added cgsnapshots and volumes relationships
    # Version 1.2: Changed 'status' field to use ConsistencyGroupStatusField
    # Version 1.3: Added cluster fields
    # Version 1.4: Added from_group
    VERSION = '1.4'

    OPTIONAL_FIELDS = ('cgsnapshots', 'volumes', 'cluster')

    fields = {
        'id': fields.UUIDField(),
        'user_id': fields.StringField(),
        'project_id': fields.StringField(),
        'cluster_name': fields.StringField(nullable=True),
        'cluster': fields.ObjectField('Cluster', nullable=True,
                                      read_only=True),
        'host': fields.StringField(nullable=True),
        'availability_zone': fields.StringField(nullable=True),
        'name': fields.StringField(nullable=True),
        'description': fields.StringField(nullable=True),
        'volume_type_id': fields.StringField(nullable=True),
        'status': c_fields.ConsistencyGroupStatusField(nullable=True),
        'cgsnapshot_id': fields.UUIDField(nullable=True),
        'source_cgid': fields.UUIDField(nullable=True),
        'cgsnapshots': fields.ObjectField('CGSnapshotList', nullable=True),
        'volumes': fields.ObjectField('VolumeList', nullable=True),
    }

    @classmethod
    def _from_db_object(cls, context, consistencygroup, db_consistencygroup,
                        expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = []
        for name, field in consistencygroup.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_consistencygroup.get(name)
            setattr(consistencygroup, name, value)

        if 'cgsnapshots' in expected_attrs:
            cgsnapshots = base.obj_make_list(
                context, objects.CGSnapshotList(context),
                objects.CGSnapshot,
                db_consistencygroup['cgsnapshots'])
            consistencygroup.cgsnapshots = cgsnapshots

        if 'volumes' in expected_attrs:
            volumes = base.obj_make_list(
                context, objects.VolumeList(context),
                objects.Volume,
                db_consistencygroup['volumes'])
            consistencygroup.volumes = volumes

        if 'cluster' in expected_attrs:
            db_cluster = db_consistencygroup.get('cluster')
            # If this consistency group doesn't belong to a cluster the cluster
            # field in the ORM instance will have value of None.
            if db_cluster:
                consistencygroup.cluster = objects.Cluster(context)
                objects.Cluster._from_db_object(context,
                                                consistencygroup.cluster,
                                                db_cluster)
            else:
                consistencygroup.cluster = None

        consistencygroup._context = context
        consistencygroup.obj_reset_changes()
        return consistencygroup

    def create(self, cg_snap_id=None, cg_id=None):
        """Create a consistency group.

        If cg_snap_id or cg_id are specified then volume_type_id,
        availability_zone, and host will be taken from the source Consistency
        Group.
        """
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already_created'))
        updates = self.cinder_obj_get_changes()

        if 'cgsnapshots' in updates:
            raise exception.ObjectActionError(action='create',
                                              reason=_('cgsnapshots assigned'))

        if 'volumes' in updates:
            raise exception.ObjectActionError(action='create',
                                              reason=_('volumes assigned'))

        if 'cluster' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('cluster assigned'))

        db_consistencygroups = db.consistencygroup_create(self._context,
                                                          updates,
                                                          cg_snap_id,
                                                          cg_id)
        self._from_db_object(self._context, self, db_consistencygroups)

    def from_group(self, group):
        """Convert a generic volume group object to a cg object."""
        self.id = group.id
        self.user_id = group.user_id
        self.project_id = group.project_id
        self.cluster_name = group.cluster_name
        self.host = group.host
        self.availability_zone = group.availability_zone
        self.name = group.name
        self.description = group.description
        self.volume_type_id = ""
        for v_type in group.volume_types:
            self.volume_type_id += v_type.id + ","
        self.status = group.status
        self.cgsnapshot_id = group.group_snapshot_id
        self.source_cgid = group.source_group_id

    def obj_load_attr(self, attrname):
        if attrname not in self.OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'cgsnapshots':
            self.cgsnapshots = objects.CGSnapshotList.get_all_by_group(
                self._context, self.id)

        if attrname == 'volumes':
            self.volumes = objects.VolumeList.get_all_by_group(self._context,
                                                               self.id)

        # If this consistency group doesn't belong to a cluster (cluster_name
        # is empty), then cluster field will be None.
        if attrname == 'cluster':
            if self.cluster_name:
                self.cluster = objects.Cluster.get_by_id(
                    self._context, name=self.cluster_name)
            else:
                self.cluster = None

        self.obj_reset_changes(fields=[attrname])

    def save(self):
        updates = self.cinder_obj_get_changes()
        if updates:
            if 'cgsnapshots' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('cgsnapshots changed'))
            if 'volumes' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('volumes changed'))
            if 'cluster' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('cluster changed'))

            db.consistencygroup_update(self._context, self.id, updates)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            updated_values = db.consistencygroup_destroy(self._context,
                                                         self.id)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())


@base.CinderObjectRegistry.register
class ConsistencyGroupList(base.ObjectListBase, base.CinderObject):
    # Version 1.0: Initial version
    # Version 1.1: Add pagination support to consistency group
    VERSION = '1.1'

    fields = {
        'objects': fields.ListOfObjectsField('ConsistencyGroup')
    }

    @staticmethod
    def include_in_cluster(context, cluster, partial_rename=True, **filters):
        """Include all consistency groups matching the filters into a cluster.

        When partial_rename is set we will not set the cluster_name with
        cluster parameter value directly, we'll replace provided cluster_name
        or host filter value with cluster instead.

        This is useful when we want to replace just the cluster name but leave
        the backend and pool information as it is.  If we are using
        cluster_name to filter, we'll use that same DB field to replace the
        cluster value and leave the rest as it is.  Likewise if we use the host
        to filter.

        Returns the number of consistency groups that have been changed.
        """
        return db.consistencygroup_include_in_cluster(context, cluster,
                                                      partial_rename,
                                                      **filters)

    @classmethod
    def get_all(cls, context, filters=None, marker=None, limit=None,
                offset=None, sort_keys=None, sort_dirs=None):
        consistencygroups = db.consistencygroup_get_all(
            context, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context),
                                  objects.ConsistencyGroup,
                                  consistencygroups)

    @classmethod
    def get_all_by_project(cls, context, project_id, filters=None, marker=None,
                           limit=None, offset=None, sort_keys=None,
                           sort_dirs=None):
        consistencygroups = db.consistencygroup_get_all_by_project(
            context, project_id, filters=filters, marker=marker, limit=limit,
            offset=offset, sort_keys=sort_keys, sort_dirs=sort_dirs)
        return base.obj_make_list(context, cls(context),
                                  objects.ConsistencyGroup,
                                  consistencygroups)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Mock unit tests for NetApp Data ONTAP FibreChannel storage systems."""

from unittest import mock

from cinder import context
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.netapp.dataontap import fakes as fake
import cinder.tests.unit.volume.drivers.netapp.fakes as na_fakes
from cinder.volume.drivers.netapp.dataontap import fc_cmode


class NetAppCmodeFibreChannelDriverTestCase(test.TestCase):

    def setUp(self):
        super(NetAppCmodeFibreChannelDriverTestCase, self).setUp()

        kwargs = {
            'configuration': self.get_config_base(),
            'host': 'openstack@netappblock',
        }
        self.library = fc_cmode.NetAppCmodeFibreChannelDriver(**kwargs)
        self.library.zapi_client = mock.Mock()
        self.zapi_client = self.library.zapi_client
        self.mock_request = mock.Mock()
        self.ctxt = context.RequestContext('fake', 'fake', auth_token=True)

    def get_config_base(self):
        return na_fakes.create_configuration()

    def test_revert_to_snapshot(self):
        mock_revert_to_snapshot = self.mock_object(self.library.library,
                                                   'revert_to_snapshot')

        self.library.revert_to_snapshot(self.ctxt, fake.SNAPSHOT_VOLUME,
                                        fake.SNAPSHOT)

        mock_revert_to_snapshot.assert_called_once_with(fake.SNAPSHOT_VOLUME,
                                                        fake.SNAPSHOT)
